# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Various helpers. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import inspect
from typing import Any, Dict, Text

# 3rd party import
import attr
from six import iteritems     # pylint: disable=wrong-import-order


@attr.s
class QsFilter(object):
    """ A halper class for parsing query string into db filters. """
    name = attr.ib(type=str)
    op = attr.ib(type=str)
    op_fn = attr.ib(type=str)
    value = attr.ib(type=Any)

    @classmethod
    def build(cls, op_map, param_name, param_value):
        # type: (Dict[Text, Text], Text, Any) -> QsFilter
        """

        Args:
            op_map (dict[str, str]):
                Operator map. This maps the query string op to a method that
                will implement this operator on the model.
            param_name (str):
                The field name as passed in the query string parameter.
            param_value (Any):
                The filter value as passed in the query string parameter.

        Returns:
            QsFilter:
                A `QsFilter` instance that can be then called on model class to
                generate the actual query filter.

        Example:

            >>> from restible.util import QsFilter
            >>> # ?name=John&age__gt=18
            >>> params = {'name': 'John', 'age__gt': 18}
            >>> OP_MAP = {
            ...     'eq': '__eq__',
            ...     'ne': '__ne__',
            ...     'lt': '__lt__',
            ...     'gt': '__gt__',
            ...     'le': '__le__',
            ...     'ge': '__ge__',
            ...     'in': 'in_',
            ... }
            >>>
            >>> filters = [
            ...     QsFilter.build(OP_MAP, k, v) for k, v in params.items()
            ... ]

        """
        parts = param_name.rsplit('__', 1)
        if len(parts) == 2:
            name, op = parts
        else:
            op = 'eq'
            name = parts[0]

        rv = cls(name=name, op=op, op_fn=op_map.get(op), value=param_value)

        if not rv.op_fn:
            msg = "Invalid OP: {0.op} in '{0.name} {0.op} {0.value}'"
            raise ValueError(msg.format(rv))

        return rv

    def __call__(self, model_cls):
        field = getattr(model_cls, self.name)
        op_method = getattr(field, self.op_fn)
        return op_method(self.value)


def iter_public_props(obj, predicate=None):
    """ Iterate over public properties of an object.

    :param Any obj:
        The object we want to get the properties of.
    :param function predicate:
        Additional predicate to filter out properties we're interested in. This
        function will be called on every property of the object with the
        property name and value as it arguments. If it returns True, the
        property will be yielded by this generator.
    """
    predicate = predicate or (lambda n, v: True)
    obj_type = type(obj)

    if inspect.isclass(obj):
        # This is a class
        for name, value in iteritems(obj.__dict__):
            if isinstance(value, property):
                yield name, value
    else:
        # This is an instance
        for name in dir(obj):
            if name.startswith('_'):
                continue

            member = getattr(obj_type, name)
            if not isinstance(member, property):
                continue

            try:
                value = getattr(obj, name)
                if predicate(name, value):
                    yield name, value
            except AttributeError:
                pass


def update_from_values(obj, values):
    # type: (object, dict) -> None
    """ Update object attributes from a dict of values.

    This is a small helper to update any object with values from a dictionary.

    Args:
        obj (object):
        values (dict[str, Any]):

    Examples:

        >>> from restible import util
        >>>
        >>> class Person(object):
        ...      def __init__(self):
        ...         self.name = 'John'
        ...         self.age = 35
        >>>
        >>> person = Person()
        >>> util.update_from_values(person, {
        ...     'name': 'Dave',
        ...     'age': 28,
        ... })
        >>> person.name
        'Dave'
        >>> person.age
        28

    """
    for name, value in iteritems(values):
        setattr(obj, name, value)


# Used only in type hint comments
del Dict, Text



# -*- coding: utf-8 -*-
""" Exception classes for restible. """
from __future__ import absolute_import, unicode_literals


class Error(Exception):
    """ Base exception class for restible related errors. """
    status = 500
    detail = "Unknown Error"

    def __init__(self, detail=None, status=None):
        super(Error, self).__init__(detail)
        self.detail = detail or self.detail
        self.status = status or self.status


class BadRequest(Error):
    """ Will be converted to HTTP 400 Bad Request. """
    status = 400
    detail = "Bad Request"


class NotAuthorized(Error):
    """ Will be converted to HTTP 401 Not Authorized. """
    status = 401
    detail = "Not Authorized"


class NotAllowed(Error):
    """ Will be converted to HTTP 403 Not Allowed. """
    status = 403
    detail = "Not Allowed"


class NotFound(Error):
    """ Will be converted to HTTP 404 Not Found. """
    status = 404
    detail = "Not Found"



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Functionality related to filters.

Filters are used by when querying the resource. For now the functionality is
quite simple - we just extract the filters from GET query params.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re
from datetime import datetime


# YYYY-mm-dd                    2018-09-25
re_date = re.compile(r'\d{4}-\d{2}-\d{2}')
# YYYY-mm-ddTHH:MM:SS           2018-09-25T19:34:55
re_datetime = re.compile(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}\d{2}')


def parse(get_params):
    """ Extract filters from request GET params.

    :param HttpRequest request:
    :return dict:
    """
    filters = {}
    for name, value in get_params.items():
        filters[name] = from_string(value)

    return filters


def from_string(value):
    """ Convert the given string value to the actual type it holds.

    If the string is an integer it will return an int, if it's a float, then it
    will return float otherwise it will just return a string.

    TODO: Add date and datetime support
    """
    try:
        if re_date.match(value):
            return datetime.strptime(value, '%Y-%m-%d').date()
        elif re_datetime.match(value):
            return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
        if value.isdigit():
            return int(value)
        else:
            return float(value)

    except ValueError:
        return value



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Simple REST API implementation.

Each endpoint has it's resource class that represents all the operations
possible in that endpoint. The resource implementation should be framework
agnostic so it can be used both on AppEngine and regular django (possibly other
frameworks as well..
"""
from __future__ import absolute_import, unicode_literals

# package interface
from .endpoint import RestEndpoint, RawResponse
from .exc import Error
from .exc import BadRequest
from .exc import NotAllowed
from .exc import NotAuthorized
from .exc import NotFound
from .model import ModelResource
from .resource import RestResource
from .actions import api_action
from .routing import api_route

__version__ = '0.11.7'

__all__ = [
    'api_action',
    'api_route',
    'ModelResource',
    'RawResponse',
    'RestEndpoint',
    'RestResource',
]
